@@ -132,18 +132,46 @@ class ActionClassifier(Module):
     def action_variance(self):
         return einx.divide('b d, b', self.action_sum_diff_squared, self.action_counts - 1)
 
+    def get_stats_from_dataset_(
+        self,
+        dataset,
+        batch_size = 32,
+        parallel = True
+    ):
+        dataloader = DataLoader(dataset, batch_size = batch_size)
+
+        update_stats_fn = self.update_action_statistics_with_parallel_welford_ if parallel else self.update_action_statistics_with_welford_
+
+        for actions, action_types in dataloader:
+            update_stats_fn(actions, action_types)
+
+    def standardize_shapes(
+        self,
+        actions,
+        action_types
+    ):
+        if actions.ndim == 3:
+            times = actions.shape[1]
+            actions = rearrange(actions, 'b t d -> (b t) d')
+
+            # the entire time chunk is of one action type
+
+            if action_types.ndim == 1:
+                action_types = repeat(action_types, 'b -> (b t)', t = times)
+
+        if action_types.ndim == 2:
+            action_types = rearrange(action_types, 'b t -> (b t)')
+
+        assert actions.shape[0] == action_types.shape[0]
+
+        return actions, action_types
+
     def update_action_statistics_with_welford_(
         self,
         actions,      # (b d) | (b t d)
         action_types  # (b) | (b t)
     ):
-        if actions.ndim == 3:
-            times = actions.shape[1]
-            actions = rearrange(actions, 'b t d -> (b t) d')
-            action_types = repeat(action_types, 'b -> (b t)', t = times)
-
-        if action_types.ndim == 2:
-            action_types = rearrange(action_types, 'b t -> (b t)')
+        actions, action_types = self.standardize_shapes(actions, action_types)
 
         for one_action, action_type in zip(actions, action_types):
 
@@ -165,13 +193,7 @@ class ActionClassifier(Module):
         actions,      # (b d) | (b t d)
         action_types  # (b) | (b t)
     ):
-        if actions.ndim == 3:
-            times = actions.shape[1]
-            actions = rearrange(actions, 'b t d -> (b t) d')
-            action_types = repeat(action_types, 'b -> (b t)', t = times)
-
-        if action_types.ndim == 2:
-            action_types = rearrange(action_types, 'b t -> (b t)')
+        actions, action_types = self.standardize_shapes(actions, action_types)
 
         batch, device = actions.shape[0], actions.device
 
@@ -239,7 +261,7 @@ class ActionClassifier(Module):
         inv_std = einx.where('b,, b d', counts == 0, 1., inv_std)
 
         mean_centered = einx.subtract('b t d, b d', actions, means)
-        normed = einx.multiply('b t d, b d', actions, inv_std)
+        normed = einx.multiply('b t d, b d', mean_centered, inv_std)
         return normed
 
     def inverse_normalize(
@@ -400,6 +422,7 @@ class LBM(Module):
         self,
         action_dim,
         dim_pose,
+        action_chunk_size = None,
         action_chunk_normalizer: ActionClassifier | None = None,
         dim = 768,
         depth = 8, # Table 2. - not very deep at all
@@ -523,6 +546,9 @@ class LBM(Module):
 
         # optional action normalizer - needs to be pretrained if passed in
 
+        assert not (exists(action_chunk_normalizer) and exists(action_mean_std_for_norm))
+
+        self.action_chunk_size = action_chunk_size
         self.action_chunk_normalizer = action_chunk_normalizer
         self.normalize_with_action_classifier = exists(action_chunk_normalizer)
 
@@ -660,6 +686,22 @@ class LBM(Module):
 
         sampled_actions, noise =  self.gaussian_diffusion_1d.sample(batch_size = batch_size, return_noise = True, model_forward_kwargs = model_forward_kwargs)
 
+        if self.normalize_with_action_classifier:
+            action_len = sampled_actions.shape[1]
+            needs_chunking = exists(self.action_chunk_size) and action_len > self.action_chunk_size
+
+            if needs_chunking:
+                assert divisible_by(action_len, self.action_chunk_size)
+
+                sampled_actions = rearrange(sampled_actions, 'b (c t) ... -> b c t ...')
+                sampled_actions, packed_shape = pack([sampled_actions], '* t d')
+
+            pred_action_types, sampled_actions = self.action_chunk_normalizer(sampled_actions)
+
+            if needs_chunking:
+                sampled_actions, = unpack([sampled_actions], packed_shape, '* t d')
+                sampled_actions = rearrange(sampled_actions, 'b c t ... -> b (c t) ...')
+
         if self.add_task_status_prediction and remove_task_status:
             # remove task status during inference
             # todo - should consider also fixing it at 0 and infill
@@ -671,9 +713,6 @@ class LBM(Module):
             mean, std = self.action_mean_std_for_norm.unbind(dim = -1)
             sampled = sampled * std + mean
 
-        if self.normalize_with_action_classifier:
-            sampled_actions, pred_action_types = self.action_chunk_normalizer(sampled_actions)
-
         if not return_noise:
             return sampled_actions
 
@@ -755,7 +794,30 @@ class LBM(Module):
         assert xnor(self.normalize_with_action_classifier, exists(action_types))
 
         if self.normalize_with_action_classifier:
-            actions = self.action_chunk_normalizer(actions)
+
+            action_len = actions.shape[1]
+            needs_chunking = exists(self.action_chunk_size) and action_len > self.action_chunk_size
+
+            # if action being trained on is multiple chunks
+
+            if needs_chunking:
+                assert divisible_by(action_len, self.action_chunk_size)
+
+                if action_types.ndim == 1:
+                    action_types = rearrange(action_types, 'b -> b 1')
+
+                actions = rearrange(actions, 'b (c t) d -> b c t d', t = self.action_chunk_size)
+                action_types = repeat(action_types, 'b c -> b (c r)', r = action_len // self.action_chunk_size)
+
+                actions, action_types = tuple(rearrange(t, 'b c ... -> (b c) ...') for t in (actions, action_types))
+
+            # normalize actions
+
+            actions = self.action_chunk_normalizer.normalize(actions, action_types)
+
+            if needs_chunking:
+                actions, = unpack(actions, '* t d')
+                actions = rearrange(actions, 'b ... d -> b (...) d')
 
         # diffusion loss
 
